import sys 
from twisted.internet import reactor, defer, protocol, serialport
import vlabif, random
from config import *
from constants import *

#from x import OcTest

SKIP_REPROGRAM = 0
BUS_LATENCY = 11
TABLE_SIZE = 1 << ( REF_LEFT + 1 )
PERM_BITS = SP_LEFT + 1 - (CACHE_LINE_BITS + CACHE_PAGE_BITS)
NUM_PAGES = PERM_SIZE = 1 << PERM_BITS
CACHE_LINE_SIZE = 1 << CACHE_LINE_BITS
HT = 3


[ BUS_INACTIVE, 
BUS_WRITE_DELAY, BUS_READ_DELAY, BUS_WRITE,
BUS_READ, BUS_FINISH ] = "iwrWRf"



@defer.inlineCallbacks
def Run():
    print 'Loading configuration'
    debug_config = vlabif.DebugConfig(dc_name="Autogen_Debug_Entity",
                chain_config_file="debug_chain.py")
    
    print 'Creating drivers'
    vlihp = vlabif.VlabInterfaceProtocol(debug=False)
    dbg = vlabif.DebugDriver(debug_config, debug=False)

    print 'Connecting to debug hardware (vlabifhw)'
    if ( not VIRTUAL_LAB ):
        # Connect via serial port
        serialport.SerialPort(vlihp, SERIAL_PORT,
                    reactor=reactor, baudrate=115200)
    else:
        # Connect via virtual lab
        import vlab
        auth = vlab.loadAuthorisation(VL_AUTH_FILE)
        vlf = vlab.VlabClientFactory(auth)
        reactor.connectTCP(auth.relay_server_name, 22, vlf)
        vl = yield vlf.getChannel()

        yield vl.connect(VL_BOARD_NAME)
        if not SKIP_REPROGRAM:
            bid = yield vl.sendBitfile(file(BIT_FILE, 'rb').read())
            print 'Bitfile sent (%d)' % bid
            rc = yield vl.programFPGA(0, bid)
        yield vl.setUART(VL_UART_NR, 115200 * 4)
        yield vl.openUART(VL_UART_NR, vlihp)
        print 'Bitfile programmed'

    yield vlihp.start()
    vlihp.openChannel(2, dbg)

    yield dbg.reset()
    yield dbg.clock(100)
    #yield dbg.setControlLines(clock = 1)
    yield dbg.capture()
    data = yield dbg.downloadChain()

    print 'Debugging hardware is responding'

    st = SMMU_Test(dbg, data)
    st.bus_latency = 10
    #yield Old_Tests2(st)

    tag_1 = 1 << (SP_LEFT + 1)
    ensure = { 0x4 | tag_1 : 0x123456 }
    store_time = yield st.Store(0x3 | tag_1, 0x123456)
    print 'write in %d' % store_time

    print 'PROGRAMMING CACHE CONTROL'
    fwd_remaps = { 0 : 0x4, 1 : 0x2 }
    z = yield st.Cache_Remap(0x0, 0x8)
    print z
    z = yield st.Cache_Remap(0x1, 0x2)
    print
    yield st.Cache_Enable(True)

    st.track_spad = False
    st.verbose = True
    st.print_chain = False



    for j in [ tag_1, 0, tag_1 ]:

        print 'Start operation at %x (cache miss expected)' % j
        (load_time, value) = yield st.Load(0x3 | j)
        print 'readback (CMx) 0x%x in %d' % (value, load_time)
        assert load_time > 10 
        assert value == st.Peek(0x3 | j)


        wval = 0xee1abc ^ j
        for i in range(CACHE_LINE_SIZE) + [3, 4, 5, 4]:
            i %= CACHE_LINE_SIZE
            print 'Start operation (cache hit expected)'
            (load_time, value) = yield st.Load(i | j)
            print 'readback (CH%d) 0x%x in %d' % (i, value, load_time)
            assert load_time <= HT
            assert value == st.Peek(i | j)
            if ensure.has_key(i | j):
                assert ensure[ i | j ] == value

            st.track_spad = True
            st.verbose = True
            st.print_chain = True

            if i == 4:
                store_time = yield st.Store(0x4 | j, wval)
                print 'write in %d' % store_time
                ensure[ i | j ] = wval
                wval += 1

        for k in xrange(2):
            i = CACHE_LINE_SIZE
            print 'Start operation (cache miss expected) k = %d' % k
            (load_time, value) = yield st.Load(i | j)
            print 'readback (CMy) 0x%x in %d' % (value, load_time)

            if (load_time <= 10):
                print 'Load time is too small!'

            if (value != st.Peek(i | j)):
                print 'value is wrong (should be 0x%x)' % st.Peek(i | j)

            assert load_time > 10 
            assert value == st.Peek(i | j)

            for i in [CACHE_LINE_SIZE + 3, 0, CACHE_LINE_SIZE - 1, 
                        (CACHE_LINE_SIZE * 2) - 1]:
                print 'Start operation (cache hit expected)'
                (load_time, value) = yield st.Load(i | j)
                print 'readback (CH%d) 0x%x in %d' % (i, value, load_time)
                assert load_time <= HT
                assert value == st.Peek(i | j)

            if k == 0:
                z = yield st.Cache_Invalidate(j | CACHE_LINE_SIZE)
                print 'invalidate %d' % z


    # This dict() has key: line_bits  and value: tag_bits
    cache_contents = { 0 : 1, 1 : 1, }

    rng = random.Random(hash('Chewy is chocolate labrador retriever'))
    used_space = set()
    for (i, j) in fwd_remaps.iteritems():
        used_space.add(j)
    unused_space = [
        j for j in xrange(PERM_SIZE) if not j in used_space ]
    rng.shuffle(unused_space)

    actual_size = PERM_SIZE / 2

    print 'Remapping remaining space'
    for i in xrange(actual_size):
        z = 0
        if not fwd_remaps.has_key(i):
            j = unused_space.pop()
            fwd_remaps[ i ] = j
            assert not j in used_space
            used_space.add(j)
            z = yield st.Cache_Remap(i, j)

        j = fwd_remaps[ i ]
        print '   %02x -> %03x (%d)' % (i, j << CACHE_LINE_BITS, z) 

    assert len(unused_space) == (PERM_SIZE - actual_size)
    rev_remaps = dict()
    for (i, j) in fwd_remaps.iteritems():
        assert not rev_remaps.has_key(j)
        rev_remaps[ j ] = i

    i = 0
    while True:
        i += 1
        fail = False
        # Hit tests are more likely for i >= 40.
        for try_again in xrange(3):
            # Choose to ensure both hit & miss test:
            tag_bits = rng.randrange(0, 3)
            line_bits = rng.randrange(0, actual_size)
            word_bits = rng.randrange(0, 1 << CACHE_LINE_BITS)
            address = word_bits | (line_bits << CACHE_LINE_BITS)
            address |= tag_bits << (SP_LEFT + 1)
            cur_tag = cache_contents.get(line_bits, None)
            if (i < 40) or (cur_tag == tag_bits):
                # Hit, or early operation
                break

        special = rng.randrange(0, 6)
        if special == 0:
            # Write something
            print 'Write test %d @ 0x%x' % (i, address),
            if cur_tag == tag_bits:
                print '(writethrough)'
            else:
                print '(no cache)'

            value = rng.randrange(0, 1 << 32)
            store_time = yield st.Store(address, value)
            check = st.Peek(address)
            if value != check:
                print 'Error: memory contents 0x%08x, should be 0x%08x' % (
                            check, value)
                fail = True

        elif special == 1:
            # Invalidate a line
            line2 = rng.randrange(0, actual_size)
            z = yield st.Cache_Invalidate(line2 << CACHE_LINE_BITS)
            assert z <= HT
            print 'Invalidate line %d' % line2
            cache_contents[ line2 ] = None


        if (special < 2) and rng.randrange(0, 2) == 0:
            # Extra load test
            (load_time, value) = yield st.Load(1)
            cache_contents[ 0 ] = 0
            check = st.Peek(1)
            if value != check:
                print 'Error: cache contents 0x%08x, should be 0x%08x' % (
                            value, check)
                fail = True

        cur_tag = cache_contents.get(line_bits, None)
        print 'Read test %d @ 0x%x (tag %d line %d word %d) sp loc %03x' % (i,
                address, tag_bits, line_bits, word_bits,
                (fwd_remaps[ line_bits ] << CACHE_LINE_BITS) | word_bits),

        if cur_tag == None:
            print '(miss/empty)'
        elif cur_tag != tag_bits:
            print '(miss)'
        else:
            print '(hit)'

        (load_time, value) = yield st.Load(address)
        if (((cur_tag == tag_bits) and (load_time > HT))
        or ((cur_tag != tag_bits) and (load_time <= HT))):
            print 'Error: delay %d' % load_time
            fail = True
    
        check = st.Peek(address)
        if value != check:
            print 'Error: cache contents 0x%08x, should be 0x%08x' % (
                        value, check)
            fail = True

        cache_contents[ line_bits ] = tag_bits
        if fail:
            break

    reactor.stop()
    return


@defer.inlineCallbacks
def Old_Tests(st):

    #st.track_spad = True
    #st.verbose = True
    #st.bus_latency = 17
    #OcTest(st)
    #return

    shadow = dict()
    r = random.Random(hash('chewy face'))

    #yield Playback(st, "mbsys/cmd.txt")
    #return

    check_val_1 = 0x12345678
    check_val_2 = 0x05c00734


    print 'opened 3 as', ref_3

    (open_time, ref_2) = yield st.Open(0x6d6, 
            0x716 - 0x6d6, 0x20, read_only = True)
    print 'opened 2 as', ref_2

    store_time = yield st.Store(0x706, check_val_1)
    print 'write 706'

    close_time = yield st.Close(ref_3)
    print 'closed 3'

    (load_time, value) = yield st.Load(0x706)
    print 'value is correct?', (value == check_val_1)

    (open_time, ref_3) = yield st.Open(0x706, 
            0x70e - 0x706, 0, read_only = True)
    print 'opened 3 as', ref_3

    print 'write 706 again'
    store_time = yield st.Store(0x706, check_val_2)
    (load_time, value) = yield st.Load(0x706)
    print 'value is correct?', (value == check_val_2)

    close_time = yield st.Close(ref_3)
    print 'closed 3'
    (load_time, value) = yield st.Load(0x706)
    print 'value is correct?', (value == check_val_2)
    print 'value is original?', (value == check_val_1)
    print 'value is 0x%x' % value 
    reactor.stop()
    return


@defer.inlineCallbacks
def Old_Tests2(st):

    r = random.Random(1234)
    shadow = dict()

    #st.track_spad = True
    #st.verbose = True
    sequence = 0

    for i in xrange(3):
        (open_time, ref) = yield st.Open(0x8 + (i * 2), 0x0f + (i * 5), 
                        (0x80 * (i + 1)) + 0x8 + (i * 2), read_only = True)
        ro_ref = ref

    while st.bus_latency > 0:
        for i in xrange(19, 0, -1):
            print 'TESTING WITH LATENCY %d - TEST %d' % (st.bus_latency, i)
            start = r.randint(0, 19)
            size = 1 + (i * 2)
            #(start,size) = (3,5) # 0x12,0x11)


            for phase in "OSCT":
                if phase == "T":
                    # Test phase!
                    for j in xrange(40):
                        (load_time, smmu_out) = yield st.Load(j)
                        print 'load time = %d' % load_time
                        shadow_out = shadow.get(j, 0xffff0000 ^ j) 
                        if (shadow_out != smmu_out):
                            print 'BUS LATENCY %d' % st.bus_latency
                            print 'Test failed: address 0x%x, shadow = 0x%x, smmu = 0x%x' % (
                                j, shadow_out, smmu_out)
                            reactor.stop()
                            return

                elif phase == "O":
                    print 'start/size = 0x%x/0x%x' % (start, size)
                    (open_time, rw_ref) = yield st.Open(start, size, start)
                    print 'open time = %d' % open_time

                elif phase == "S":
                    # Store phase
                    for address in [max(0, start - 1), start, 
                                r.randint(0, 30),
                                start + size - 1, start + size]:
                        data = (0xf << 28) | (
                            r.randint(0, 1 << 12) << 16) | (
                            (sequence & 0xff) << 8) | (
                            address & 0xff )
                        print 'store %x at %x' % (data, address)
                        store_time = yield st.Store(address, data)
                        print 'store time = %d' % store_time
                        shadow[ address ] = data
                        sequence += 1

                elif phase == "C":
                    # Close phase
                    st.track_spad = True
                    st.verbose = True
                    close_time = yield st.Close(rw_ref)
                    print 'close time = %d' % close_time
                    st.track_spad = False
                    st.verbose = False

        st.bus_latency -= 1


    store_time = yield st.Store(0x10, 0xabcdef)
    store_time = yield st.Store(0x11, 0x123456)
    print 'store in %d' % (store_time)

    (load_time, load_data) = yield st.Load(0x10)
    assert load_data == 0xabcdef
    print 'load in %d' % (load_time)

    sz = 0x10
    (open_time, rw_ref) = yield st.Open(0xa, 0x0b + sz, 0x1ca)
    print 'open ref %d delay %d' % (rw_ref, open_time)

    (load_time, load_data) = yield st.Load(0x11)
    assert load_data == 0x123456, hex(load_data)
    print 'load in %d (open_ro)' % (load_time)

    st.track_spad = True
    st.verbose = True
    store_time = yield st.Store(0x10, 0xbcdef0)
    print 'store in %d (open)' % (store_time)
    store_time = yield st.Store(0x11, 0x654321)
    print 'store in %d (open)' % (store_time)
    store_time = yield st.Store(0x9, 0xd00d5)
    print 'store in %d (open_ro)' % (store_time)

    (load_time, load_data) = yield st.Load(0x10)
    assert load_data == 0xbcdef0, hex(load_data)
    print 'load in %d (open)' % (load_time)

    (load_time, load_data) = yield st.Load(0x09)
    assert load_data == 0xd00d5, hex(load_data)
    print 'load in %d (not open)' % (load_time)

    st.verbose = True
    st.track_spad = True
    print 'BEGIN RW CLOSE'
    close_time = yield st.Close(rw_ref)
    print 'END RW CLOSE'
    print 'close ref %d delay %d' % (rw_ref, close_time)
    print 'BEGIN RO CLOSE'
    close_time = yield st.Close(ro_ref)
    print 'END RO CLOSE'
    print 'close ref %d delay %d' % (ro_ref, close_time)

    (load_time, load_data1) = yield st.Load(0x11)
    print 'load in %d (not open)' % (load_time)

    (load_time, load_data2) = yield st.Load(0x10)
    print 'load in %d (not open)' % (load_time)

    (load_time, load_data3) = yield st.Load(0x09)
    print 'load in %d (not open)' % (load_time)

    assert load_data1 == 0x654321, hex(load_data1)
    assert load_data2 == 0xbcdef0, hex(load_data2)
    assert load_data3 == 0xd00d5, hex(load_data3)

    #for i in xrange(5):
    #    base = 0x1000 + (i * 4)
    #    (open_time, ref) = yield st.Open(base, 0x10, 0x10 * i)
    #    print 'open %d in %d [%d - %d]' % (ref, open_time, base, base + 0x10)
    #    base = 0x1004 + (i * 12)
    #    store_time = yield st.Store(base, 0xabc0000 | i)
    #    print 'store in %d   [%d]' % (store_time, base)

    #store_time = yield st.Store(0x1004 + (i * 8), 0xabc0000 | i)
    #print 'store in %d' % (store_time)

    #print 'open', (open_time, ref)
    #(load1_time, load1_data) = yield st.Load(0x1234)
    #print 'load1_time', load1_time
    #print 'load1_data', hex(load1_data)
    #close_time = yield st.Close(ref)
    #print 'close', close_time

    reactor.stop()
    return 

@defer.inlineCallbacks
def Playback(st, fname):
    st.bus_latency = 16
    region = dict()
    limit = 0
    for line in file(fname):
        fields = line.split()
        if len(fields) < 4:
            continue

        line = line.rstrip()
        command = fields[ 1 ]

        if command == "OPEN":
            print line
            address = int(fields[ 2 ], 16)
            size = int(fields[ 4 ], 16)
            copy_loc = int(fields[ 6 ], 16)
            expect_ref = int(fields[ 10 ], 16)
            (open_time, ref) = yield st.Open(address, size, copy_loc)
            if ref != expect_ref:
                print 'expected ref: %d' % expect_ref
                print 'actual ref: %d' % actual_ref
                return

            limit = max(address + size, limit)
            region[ ref ] = (address, address + size)

        elif command == "CLOSE":
            print line
            ref = int(fields[ 3 ], 16)
            close_time = yield st.Close(ref)
            del region[ ref ]
           
    print 'checking to limit 0x%x' % limit
    should_be_open = set()
    todo = []
    for (ref, (begin, end)) in region.iteritems():
        print 'ref %d holds 0x%x to 0x%x' % (ref, begin, end)
        for address in xrange(begin, end, 1):
            should_be_open.add(address)

        todo.append(begin - 1)
        todo.append(begin)
        todo.append(end - 1)
        todo.append(end)
        todo.append(begin & 0x3ff)
        todo.append(begin & 0x1ff)

    todo.extend(range(0, limit, 1))
    done = set()
    error = 0
    for address in todo:
        if address in done:
            continue

        done.add(address)
        print 'Testing address 0x%x' % address
        (load_time, data) = yield st.Load(address)

        if address in should_be_open:
            if load_time > HT:
                print 'Address 0x%x should be open, but load time was %d' % (
                        address, load_time)
                error += 1
        else:
            if load_time <= HT:
                print 'Address 0x%x should not be open, but load time was %d' % (
                        address, load_time)
                error += 1

        test = 0xffff0000 ^ address
        if data != test:
            print 'For address 0x%x, data should be 0x%x, but is 0x%x' % (
                        address, test, data)
            error += 1

        if error > 5:
            print 'STOP - ERRORS FOUND'
            return

    print 'done checking'


class SMMU_Test:
    def __init__(self, dbg, data):
        self.dbg = dbg
        self.data = data
        self.memory = dict()
        self.clock = 0
        self.verbose = False
        self.track_spad = False
        self.print_chain = False
        self.data[ 'cpu_ben' ].setOutput(0xf)
        self.bus_latency = 20

    def Cache_Remap(self, src, dest):
        assert 0 <= src < PERM_SIZE
        assert 0 <= dest < PERM_SIZE
        self.data[ "cpu_address" ].setOutput(src << (
                    CACHE_LINE_BITS + CACHE_PAGE_BITS + 2))
        self.data[ "cpu_data_in" ].setOutput(dest)
        self.data[ 'cpu_cache_set_perm' ].setOutput(1)
        start = self.clock

        def GT(*args):
            return self.clock - start

        return self.Service().addCallback(GT)

    def Cache_Invalidate(self, address):
        self.data[ "cpu_address" ].setOutput(address * 4)
        self.data[ 'cpu_cache_invalidate' ].setOutput(1)
        start = self.clock

        def GT(*args):
            return self.clock - start

        return self.Service().addCallback(GT)

    @defer.inlineCallbacks
    def Cache_Enable(self, enable):
        self.data[ 'cache_active' ].setOutput(int(enable))
        yield self.dbg.uploadChain()
        yield self.dbg.ready()

    def Load(self, address):
        self.data[ 'cpu_address' ].setOutput(address * 4)
        self.data[ 'cpu_read' ].setOutput(1)
        start = self.clock

        def GDOT(*args):
            return (self.clock - start, self.data[ 'cpu_data_out' ].getInt())

        return self.Service().addCallback(GDOT)

    def Store(self, address, data):
        self.data[ 'cpu_address' ].setOutput(address * 4)
        self.data[ 'cpu_data_in' ].setOutput(data)
        self.data[ 'cpu_write' ].setOutput(1)
        start = self.clock

        def GT(*args):
            return self.clock - start

        return self.Service().addCallback(GT)

    def Open(self, address, size, copy_loc, read_only = False):
        self.data[ 'cpu_address' ].setOutput(address * 4)
        self.data[ 'cpu_copy_loc' ].setOutput(copy_loc)
        self.data[ 'cpu_size' ].setOutput(size)
        self.data[ 'cpu_open' ].setOutput(1)
        self.data[ 'cpu_read_only' ].setOutput(int(read_only))
        start = self.clock

        def GDOT(*args):
            ref = self.data[ 'cpu_data_out' ].getInt()
            assert ref < TABLE_SIZE, ref
            return (self.clock - start, ref)

        return self.Service().addCallback(GDOT)

    def Close(self, ref):
        assert ref < TABLE_SIZE, ref
        self.data[ 'cpu_data_in' ].setOutput(ref)
        self.data[ 'cpu_close' ].setOutput(1)
        start = self.clock

        def GT(*args):
            return self.clock - start

        return self.Service().addCallback(GT)

    def Peek(self, addr):
        default = 0xffff0000 ^ addr
        return self.memory.get(addr, default)

    def Poke(self, addr, data):
        self.memory[ addr ] = data

    @defer.inlineCallbacks
    def Service(self):
        # Service bus requests until 'cpu_ack'
        dbg = self.dbg
        data = self.data
        yield dbg.uploadChain()
        yield dbg.ready()

        dirty = False
        ip2bus_addr = 0
        ip2bus_be = 0
        xfer_length = 0
        latency = 0
        allow_fast = not (self.track_spad or self.verbose)
        bus_status = BUS_INACTIVE
        timer = 0
        ack = False
        while not ack:
            delta = 1
            if (latency > 2) and allow_fast and not dirty:
                delta = latency - 2

            latency = max(0, latency - delta)
            timer += delta
            self.clock += delta
            #for i in xrange(delta):
            #    yield dbg.setControlLines(clock = 0)
            #    yield dbg.setControlLines(clock = 1)
            yield dbg.clock(delta)
            yield dbg.capture()
            yield dbg.downloadChain()
            
            if self.track_spad:
                print ''
                print ''
                print self.clock
                print 'address A: %08x   address B: %08x' % (
                        data[ 'dbg_cpu_side_address' ].getInt(),
                        data[ 'dbg_bus_side_address' ].getInt())
                print '    din A: %08x       din B: %08x' % (
                        data[ 'dbg_cpu_side_data_in' ].getInt(),
                        data[ 'dbg_bus_side_data_in' ].getInt())
                print '   dout A: %08x      dout B: %08x' % (
                        data[ 'dbg_cpu_side_data_out' ].getInt(),
                        data[ 'dbg_bus_side_data_out' ].getInt())
                print '     wr A: %-8d        wr B: %-8d' % (
                        data[ 'dbg_cpu_side_write' ].getInt(),
                        data[ 'dbg_bus_side_write' ].getInt())
                print '     en A: %-8d        en B: %-8d' % (
                        1,
                        data[ 'dbg_bus_side_cen' ].getInt())
                print 'match address: %08x' % (
                        data[ 'dbg_match_address' ].getInt())

                for extra in [ 'dbg_match', 'dbg_match_in_read_only_area',
                        'dbg_miss', 'dbg_hit', "dbg_match_ref",
                        "dbg_row_activation",
                        "dbg_code", "dbg_sp_cpu_valid", "dbg_slow_ack", 
                        'dbg_read_only_update_pos',
                        'dbg_update_burst_boundary' ]:
                    print '    %s: %u  (0x%x)' % (extra, 
                                data[ extra ].getInt(),
                                data[ extra ].getInt())

            if self.verbose:
                if data[ 'dbg_cpu_side_write' ].getInt():
                    print ('%d: cpu side: write scratchpad address '
                                '0x%x with data 0x%x' % (
                            self.clock, 
                            data[ 'dbg_cpu_side_address' ].getInt(),
                            data[ 'dbg_cpu_side_data_in' ].getInt()))
                if data[ 'dbg_bus_side_write' ].getInt():
                    print ('%d: bus side: write scratchpad address '
                                '0x%x with data 0x%x' % (
                            self.clock, 
                            data[ 'dbg_bus_side_address' ].getInt(),
                            data[ 'dbg_bus_side_data_in' ].getInt()))

            if self.print_chain:
                dbg.printChain()

            #print ''
            #print ''
            #print ''
            #print self.clock

            assert timer < 1000

            ack = data[ 'cpu_ack' ].getInt()

            capture_address = False
            dirty = False

            # Make sure that one-shot signals are zeroed
            for cancel_flag in [ 'mst_cmd_sm_clr_go',
                        'cpu_cache_set_perm',
                        'cpu_cache_invalidate',
                        'cpu_address', 'cpu_data_in',
                        'cpu_read', 'cpu_write',
                        'cpu_open', 'cpu_close',
                        'input_next', 'output_next' ]:
                if data[ cancel_flag ].getInt():
                    dirty = True
                    data[ cancel_flag ].setOutput(0)

            if data[ 'mst_go' ].getInt():
                assert bus_status == BUS_INACTIVE
                if data[ 'mst_cntl_wr_req' ].getInt():
                    bus_status = BUS_WRITE_DELAY
                    if self.verbose:
                        print 'begin write delay'
                if data[ 'mst_cntl_rd_req' ].getInt():
                    assert bus_status == BUS_INACTIVE
                    bus_status = BUS_READ_DELAY
                    if self.verbose:
                        print 'begin read delay'
                assert bus_status != BUS_INACTIVE, bus_status

                ip2bus_addr = data[ 'mst_ip2bus_addr' ].getInt() / 4
                ip2bus_be = data[ 'mst_ip2bus_be' ].getInt()
                xfer_length = data[ 'mst_xfer_length' ].getInt()
                latency = self.bus_latency
                assert (xfer_length % 4) == 0
                xfer_length /= 4
                assert 1 <= xfer_length <= 16
                assert ip2bus_addr >= 0

                dirty = True
                data[ 'mst_cmd_sm_busy' ].setOutput(1)
                data[ 'mst_cmd_sm_clr_go' ].setOutput(1)

            if bus_status in (BUS_WRITE_DELAY, BUS_READ_DELAY):
                if latency <= 0:
                    if bus_status == BUS_WRITE_DELAY:
                        bus_status = BUS_WRITE
                        if self.verbose:
                            print 'begin write'
                    else:
                        bus_status = BUS_READ
                        if self.verbose:
                            print 'begin read'

            elif bus_status == BUS_READ:
                dirty = True
                data[ 'data_input' ].setOutput(self.Peek(ip2bus_addr))
                data[ 'input_next' ].setOutput(1)
                if self.verbose:
                    print '%d: read address %x with data %x' % (
                            self.clock, ip2bus_addr,
                            data[ 'data_input' ].getInt())
                ip2bus_addr += 1
                xfer_length -= 1
                if xfer_length == 0:
                    bus_status = BUS_FINISH

            elif bus_status == BUS_WRITE:
                dirty = True
                if self.verbose:
                    print '%d: write address %x with data %x' % (
                            self.clock, ip2bus_addr,
                            data[ 'data_output' ].getInt())
                self.Poke(ip2bus_addr, data[ 'data_output' ].getInt())
                data[ 'output_next' ].setOutput(1)
                ip2bus_addr += 1
                xfer_length -= 1
                if xfer_length == 0:
                    bus_status = BUS_FINISH

            elif bus_status == BUS_FINISH:
                dirty = True
                data[ 'mst_cmd_sm_busy' ].setOutput(0)
                if self.verbose:
                    print 'completion'
                bus_status = BUS_INACTIVE

            check = self.data[ 'cpu_data_out' ].getInt()
            if (((check & 0xfffff000) == 0xe0e0e000)
            and ((check & 0xfff) != 0)): # 0 is reset condition
                print '%d: Warning: error condition 0x%x present, bs %s' % (
                        self.clock, check, bus_status)

            if dirty:
                yield dbg.uploadChain()
                yield dbg.ready()


        assert bus_status == BUS_INACTIVE, bus_status
        # Not validated:
        #  mst_cntl_burst
        #  mst_cntl_bus_lock (always 0)



if ( __name__ == "__main__" ):
    reactor.addSystemEventTrigger('after', 'startup', Run)
    reactor.run()


