# 
#  Scratchpad MMU - Virtual lab launcher for SMMU tests
#  Copyright (C) Jack Whitham 2009
#  http://www.jwhitham.org.uk/c/smmu.html
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation
#  (version 2.1 of the License only).
#  
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#  
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
# 
import vlab, sys
from twisted.internet import reactor, defer
from vlab import terminal

AUTH = "/home/jack/jack-vl2.key"

@defer.inlineCallbacks
def Run():
    if ( len(sys.argv) != 3 ):
        print "Usage: %s <code (.bin)> <FPGA (.bit)>" % sys.argv[ 0 ]
        reactor.stop()
        return

    code_fname = sys.argv[ 1 ]
    bit_fname = sys.argv[ 2 ]

    code = file(code_fname, "rb").read()
    assert ( len(code) & 3 ) == 0
    bits = file(bit_fname, 'rb').read()
    print '%u code bytes, %u FPGA bytes' % (len(code), len(bits))

    auth = vlab.loadAuthorisation(AUTH)
    vlf = vlab.VlabClientFactory(auth)
    reactor.connectTCP(auth.relay_server_name, 22, vlf)
    vl = yield vlf.getChannel()
    yield vl.connect("ml505")
   
    yield vl.setUART(1, 115200)

    print 'Sending bit file (%u bytes)' % len(bits)
    bid = yield vl.sendBitfile(bits)

    print 'Bitfile sent, bid %u' % bid
    uart = vlab.VlabUARTProtocol()
    rc = yield vl.programFPGAopenUART(0, bid, 1, uart)

    print 'Programming complete, rc %u' % rc
    print 'Wait for "hi"'
    data = yield uart.read(2)
    assert data == "hi"

    uart.write(code)
    if False:
        size = len(code)
        print 'Send size (%u)' % size
        uart.write(chr((size >> 24) & 0xff))
        uart.write(chr((size >> 16) & 0xff))
        uart.write(chr((size >> 8) & 0xff))
        uart.write(chr((size >> 0) & 0xff))
        data = yield uart.read(1)
        assert data == "l", data

        target = 0x900 << 20
        print 'Send target (0x%x)' % target
        uart.write(chr((target >> 24) & 0xff))
        uart.write(chr((target >> 16) & 0xff))
        uart.write(chr((target >> 8) & 0xff))
        uart.write(chr((target >> 0) & 0xff))
        data = yield uart.read(1)
        assert data == "k", data

        print 'Send code'
        uart.write(code)

    data = yield uart.read(1)
    i = 0
    while ( data != "g" ):
        sys.stdout.write(data)
        sys.stdout.flush()
        assert data in ".lk"
        i += 0x1000
        data = yield uart.read(1)
    assert data == "g"

    print 'Enter relay loop'
    while True:
        data = yield uart.read(1)
        sys.stdout.write(data)
        sys.stdout.flush()

if ( __name__ == "__main__" ):
    reactor.addSystemEventTrigger('after', 'startup', Run)
    reactor.run()

